@@ -4,6 +4,7 @@ import string
 import sys
 import tempfile
 from collections import defaultdict
+from enum import Enum, auto
 from pathlib import Path, PurePath
 from typing import (
     Dict,
@@ -33,6 +34,8 @@ from rasterio.enums import Resampling
 from rasterio.io import DatasetWriter, MemoryFile
 from rasterio.shutil import copy as rio_copy
 from rasterio.warp import calculate_default_transform, reproject
+from scipy.ndimage import binary_fill_holes
+from shapely.geometry import box
 from shapely.geometry.base import CAP_STYLE, JOIN_STYLE, BaseGeometry
 
 from eodatasets3.model import DatasetDoc, GridDoc, MeasurementDoc
@@ -46,6 +49,38 @@ except ImportError:
     h5py = None
 
 
+class ValidDataMethod(Enum):
+    """
+    How to calculate the valid data geometry for an image?
+    """
+
+    #: Vectorize the full valid pixel mask as-is.
+    #:
+    #: In some circumstances this can be very slow.
+    #: `filled` may be safer.
+    #:
+    thorough = auto()
+
+    #: Fill holes in the valid pixel mask before vectorizing.
+    #:
+    #: (Potentially much faster than ``thorough`` if there's many small
+    #: nodata holes, as they will create many tiny polygons.
+    #: *slightly* slower if no holes exist.)
+    filled = auto()
+
+    #: Take convex-hull of valid pixel mask before vectorizing.
+    #:
+    #: This is much slower than ``filled``, but will work in cases where
+    #: you have a lot of internal geometry that aren't holes.
+    #: Such as SLC-Off Landsat 7 data.
+    #:
+    #: Requires 'scikit-image' dependency.
+    convex_hull = auto()
+
+    #: Use the image file bounds, ignoring actual pixel values.
+    bounds = auto()
+
+
 @attr.s(auto_attribs=True, slots=True, hash=True, frozen=True)
 class GridSpec:
     """
@@ -421,60 +456,42 @@ class MeasurementBundler:
                 )
         return crs, grid_docs, measurement_docs
 
-    def consume_and_get_valid_data(self) -> BaseGeometry:
+    def consume_and_get_valid_data(
+        self, valid_data_method: ValidDataMethod = ValidDataMethod.thorough
+    ) -> BaseGeometry:
         """
         Consume the stored grids and produce the valid data for them.
 
         (they are consumed in order to to minimise peak memory usage)
+
+        :param valid_data_method: How to calculate the valid-data polygon?
         """
 
-        def valid_shape(shape):
-            if shape.is_valid:
-                return shape
-            return shape.buffer(0)
-
         geoms = []
+
         while self.mask_by_grid:
             grid, mask = self.mask_by_grid.popitem()
-            mask = mask.astype("uint8")
-            shape = shapely.ops.unary_union(
-                [
-                    valid_shape(shapely.geometry.shape(shape))
-                    for shape, val in rasterio.features.shapes(mask)
-                    if val == 1
-                ]
-            )
-            shape_y, shape_x = mask.shape
-            del mask
 
-            # convex hull
-            geom = shape.convex_hull
+            if valid_data_method is ValidDataMethod.bounds:
+                geom = box(*grid.bounds)
+            elif valid_data_method is ValidDataMethod.filled:
+                mask = mask.astype("uint8")
+                binary_fill_holes(mask, output=mask)
+                geom = _grid_to_poly(grid, mask)
+            elif valid_data_method is ValidDataMethod.convex_hull:
+                # Requires optional dependency scikit-image
+                from skimage import morphology as morph
 
-            # buffer by 1 pixel
-            geom = geom.buffer(
-                1, cap_style=CAP_STYLE.square, join_style=JOIN_STYLE.bevel
-            )
-
-            # simplify with 1 pixel radius
-            geom = geom.simplify(1)
-
-            # intersect with image bounding box
-            geom = geom.intersection(shapely.geometry.box(0, 0, shape_x, shape_y))
-
-            # transform from pixel space into CRS space
-            geom = shapely.affinity.affine_transform(
-                geom,
-                (
-                    grid.transform.a,
-                    grid.transform.b,
-                    grid.transform.d,
-                    grid.transform.e,
-                    grid.transform.xoff,
-                    grid.transform.yoff,
-                ),
-            )
+                geom = _grid_to_poly(
+                    grid, morph.convex_hull_image(mask).astype("uint8")
+                )
+            elif valid_data_method is ValidDataMethod.thorough:
+                geom = _grid_to_poly(grid, mask.astype("uint8"))
+            else:
+                raise NotImplementedError(
+                    f"Unexpected valid data method: {valid_data_method}"
+                )
             geoms.append(geom)
-
         return shapely.ops.unary_union(geoms)
 
     def iter_names(self) -> Generator[str, None, None]:
@@ -490,6 +507,45 @@ class MeasurementBundler:
                 yield grid, band_name, meas_path.path
 
 
+def _valid_shape(shape: BaseGeometry) -> BaseGeometry:
+    if shape.is_valid:
+        return shape
+    return shape.buffer(0)
+
+
+def _grid_to_poly(grid: GridSpec, mask: numpy.ndarray) -> BaseGeometry:
+    shape = shapely.ops.unary_union(
+        [
+            _valid_shape(shapely.geometry.shape(shape))
+            for shape, val in rasterio.features.shapes(mask)
+            if val == 1
+        ]
+    )
+    shape_y, shape_x = mask.shape
+    del mask
+    # convex hull
+    geom = shape.convex_hull
+    # buffer by 1 pixel
+    geom = geom.buffer(1, cap_style=CAP_STYLE.square, join_style=JOIN_STYLE.bevel)
+    # simplify with 1 pixel radius
+    geom = geom.simplify(1)
+    # intersect with image bounding box
+    geom = geom.intersection(shapely.geometry.box(0, 0, shape_x, shape_y))
+    # transform from pixel space into CRS space
+    geom = shapely.affinity.affine_transform(
+        geom,
+        (
+            grid.transform.a,
+            grid.transform.b,
+            grid.transform.d,
+            grid.transform.e,
+            grid.transform.xoff,
+            grid.transform.yoff,
+        ),
+    )
+    return geom
+
+
 @attr.s(auto_attribs=True)
 class WriteResult:
     # path: Path
